#pragma once

#include "wled.h"
#include "battery_defaults.h"

/*
 * Usermod by Maximilian Mewes
 * Mail: mewes.maximilian@gmx.de
 * GitHub: itCarl
 * Date: 25.12.2022
 * If you have any questions, please feel free to contact me.
 */
class UsermodBattery : public Usermod 
{
  private:
    // battery pin can be defined in my_config.h
    int8_t batteryPin = USERMOD_BATTERY_MEASUREMENT_PIN;
    // how often to read the battery voltage
    unsigned long readingInterval = USERMOD_BATTERY_MEASUREMENT_INTERVAL;
    unsigned long nextReadTime = 0;
    unsigned long lastReadTime = 0;
    // battery min. voltage
    float minBatteryVoltage = USERMOD_BATTERY_MIN_VOLTAGE;
    // battery max. voltage
    float maxBatteryVoltage = USERMOD_BATTERY_MAX_VOLTAGE;
    // all battery cells summed up
    unsigned int totalBatteryCapacity = USERMOD_BATTERY_TOTAL_CAPACITY;
    // raw analog reading 
    float rawValue = 0.0f;
    // calculated voltage            
    float voltage = maxBatteryVoltage;
    // between 0 and 1, to control strength of voltage smoothing filter
    float alpha = 0.05f;
    // multiplier for the voltage divider that is in place between ADC pin and battery, default will be 2 but might be adapted to readout voltages over ~5v ESP32 or ~6.6v ESP8266
    float voltageMultiplier = USERMOD_BATTERY_VOLTAGE_MULTIPLIER;
    // mapped battery level based on voltage
    int8_t batteryLevel = 100;
    // offset or calibration value to fine tune the calculated voltage
    float calibration = USERMOD_BATTERY_CALIBRATION;

    // auto shutdown/shutoff/master off feature
    bool autoOffEnabled = USERMOD_BATTERY_AUTO_OFF_ENABLED;
    int8_t autoOffThreshold = USERMOD_BATTERY_AUTO_OFF_THRESHOLD;

    // low power indicator feature
    bool lowPowerIndicatorEnabled = USERMOD_BATTERY_LOW_POWER_INDICATOR_ENABLED;
    int8_t lowPowerIndicatorPreset = USERMOD_BATTERY_LOW_POWER_INDICATOR_PRESET;
    int8_t lowPowerIndicatorThreshold = USERMOD_BATTERY_LOW_POWER_INDICATOR_THRESHOLD;
    int8_t lowPowerIndicatorReactivationThreshold = lowPowerIndicatorThreshold+10;
    int8_t lowPowerIndicatorDuration = USERMOD_BATTERY_LOW_POWER_INDICATOR_DURATION;
    bool lowPowerIndicationDone = false;
    unsigned long lowPowerActivationTime = 0; // used temporary during active time
    int8_t lastPreset = 0;

    bool initDone = false;
    bool initializing = true;

    // strings to reduce flash memory usage (used more than twice)
    static const char _name[];
    static const char _readInterval[];
    static const char _enabled[];
    static const char _threshold[];
    static const char _preset[];
    static const char _duration[];
    static const char _init[];

    // custom map function
    double mapf(double x, double in_min, double in_max, double out_min, double out_max) 
    {
      return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
    }

    float dot2round(float x) 
    {
      float nx = (int)(x * 100 + .5);
      return (float)(nx / 100);
    }

    /*
     * Turn off all leds
     */
    void turnOff()
    {
      bri = 0;
      stateUpdated(CALL_MODE_DIRECT_CHANGE);
    }

    /*
     * Indicate low power by activating a configured preset for a given time and then switching back to the preset that was selected previously
     */
    void lowPowerIndicator()
    {
      if (!lowPowerIndicatorEnabled) return;
      if (batteryPin < 0) return;
      if (lowPowerIndicationDone && lowPowerIndicatorReactivationThreshold <= batteryLevel) lowPowerIndicationDone = false;
      if (lowPowerIndicatorThreshold <= batteryLevel) return;
      if (lowPowerIndicationDone) return;
      if (lowPowerActivationTime <= 1) {
        lowPowerActivationTime = millis();
        lastPreset = currentPreset;
        applyPreset(lowPowerIndicatorPreset);
      }

      if (lowPowerActivationTime + (lowPowerIndicatorDuration * 1000) <= millis()) {
        lowPowerIndicationDone = true;
        lowPowerActivationTime = 0;
        applyPreset(lastPreset);
      }      
    }

    /*
     * Function to read battery voltage
     */
    float readVoltage() {
      #ifdef ARDUINO_ARCH_ESP32
        float measuredVoltage = (analogReadMilliVolts(batteryPin) / 1000.0f) * voltageMultiplier + calibration;
        // Přidáme kontrolní výstupy
        Serial.print("Voltage Multiplier: ");
        Serial.println(voltageMultiplier);
        Serial.print("Calibration: ");
        Serial.println(calibration);
        Serial.print("Measured Voltage: ");
        Serial.println(measuredVoltage);
        return measuredVoltage;
      #else
        float measuredVoltage = (analogRead(batteryPin) / 1023.0f) * voltageMultiplier + calibration;
        Serial.print("Voltage Multiplier: ");
        Serial.println(voltageMultiplier);
        Serial.print("Calibration: ");
        Serial.println(calibration);
        Serial.print("Measured Voltage: ");
        Serial.println(measuredVoltage);
        return measuredVoltage;
      #endif
    }

  public:
    /*
     * setup() is called once at boot. WiFi is not yet connected at this point.
     */
    void setup() 
    {
      #ifdef ARDUINO_ARCH_ESP32
        bool success = false;
        DEBUG_PRINTLN(F("Allocating battery pin..."));
        if (batteryPin >= 0 && digitalPinToAnalogChannel(batteryPin) >= 0) 
          if (pinManager.allocatePin(batteryPin, false, PinOwner::UM_Battery)) {
            DEBUG_PRINTLN(F("Battery pin allocation succeeded."));
            success = true;
            voltage = readVoltage();
          }

        if (!success) {
          DEBUG_PRINTLN(F("Battery pin allocation failed."));
          batteryPin = -1; 
        } else {
          pinMode(batteryPin, INPUT);
        }
      #else 
        pinMode(batteryPin, INPUT);
        voltage = readVoltage();
      #endif

      nextReadTime = millis() + readingInterval;
      lastReadTime = millis();

      initDone = true;
    }

    /*
     * loop() is called continuously.
     */
    void loop() 
    {
      if(strip.isUpdating()) return;

      lowPowerIndicator();

      if (millis() < nextReadTime) return;

      nextReadTime = millis() + readingInterval;
      lastReadTime = millis();

      if (batteryPin < 0) return;

      initializing = false;

      rawValue = readVoltage();
      voltage = voltage + alpha * (rawValue - voltage);

      #ifdef USERMOD_BATTERY_USE_LIPO
        batteryLevel = mapf(voltage, minBatteryVoltage, maxBatteryVoltage, 0, 100);
        if (batteryLevel < 40.0f) 
          batteryLevel = mapf(batteryLevel, 0, 40, 0, 12);      
        else if (batteryLevel < 90.0f)
            batteryLevel = mapf(batteryLevel, 40, 90, 12, 95);   
        else 
            batteryLevel = mapf(batteryLevel, 90, 105, 95, 100); 
      #else
        batteryLevel = mapf(voltage, minBatteryVoltage, maxBatteryVoltage, 0, 100);
      #endif
      if (voltage > -1.0f) batteryLevel = constrain(batteryLevel, 0.0f, 110.0f);

      if (autoOffEnabled && (autoOffThreshold >= batteryLevel))
        turnOff();
    }

    /*
     * addToJsonInfo() adds custom entries to the /json/info part of the JSON API.
     */
    void addToJsonInfo(JsonObject& root)
    {
      JsonObject user = root["u"];
      if (user.isNull()) user = root.createNestedObject("u");

      if (batteryPin < 0) {
        JsonArray infoVoltage = user.createNestedArray(F("Battery voltage"));
        infoVoltage.add(F("n/a"));
        infoVoltage.add(F(" invalid GPIO"));
        return;
      }

      JsonArray infoPercentage = user.createNestedArray(F("Battery level"));
      JsonArray infoVoltage = user.createNestedArray(F("Battery voltage"));
      JsonArray infoNextUpdate = user.createNestedArray(F("Next update"));

      infoNextUpdate.add((nextReadTime - millis()) / 1000);
      infoNextUpdate.add(F(" sec"));

      if (initializing) {
        infoPercentage.add(FPSTR(_init));
        infoVoltage.add(FPSTR(_init));
        return;
      }

      infoPercentage.add(batteryLevel);
      infoPercentage.add(F(" %"));
      infoVoltage.add(dot2round(voltage));
      infoVoltage.add(F(" V"));
    }
};

// strings to reduce flash memory usage
const char UsermodBattery::_name[]          PROGMEM = "Battery";
const char UsermodBattery::_readInterval[]  PROGMEM = "interval";
const char UsermodBattery::_enabled[]       PROGMEM = "enabled";
const char UsermodBattery::_threshold[]     PROGMEM = "threshold";
const char UsermodBattery::_preset[]        PROGMEM = "preset";
const char UsermodBattery::_duration[]      PROGMEM = "duration";
const char UsermodBattery::_init[]          PROGMEM = "init";
